/*
 * This file is a modification of an ITK source code file.
 * It uses B-spline weight caching as in itkMattesMutualInformationImageToImageMetric.
 * See below for the original header and license.
 *
 * Chris Schwemmer, Universitt Erlangen-Nrnberg
 * chris.schwemmer@cs.fau.de
 */

/*=========================================================================

  Program:   Insight Segmentation & Registration Toolkit
  Module:    $RCSfile: itkNormalizedCorrelationImageToImageMetricFast.txx,v $
  Language:  C++
  Date:      $Date: 2009-01-26 21:45:54 $
  Version:   $Revision: 1.49 $

  Copyright (c) Insight Software Consortium. All rights reserved.
  See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __itkNormalizedCorrelationImageToImageMetricFast_txx
#define __itkNormalizedCorrelationImageToImageMetricFast_txx

#include "itkNormalizedCorrelationImageToImageMetricFast.h"
#include "itkImageRegionConstIteratorWithIndex.h"

namespace itk
{

/**
 * Constructor
 */
template <class TFixedImage, class TMovingImage> 
NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::NormalizedCorrelationImageToImageMetricFast()
{
  m_SubtractMean = false;
  m_TransformIsBSpline = false;

  typename BSplineTransformType::Pointer transformer = BSplineTransformType::New();
  this->SetTransform(transformer);

  m_NumberOfSpatialSamples = 0;
  m_NumParametersPerDim = 0;
  m_NumBSplineWeights = 0;
  m_BSplineTransform = NULL;
  m_UseCachingOfBSplineWeights = true;
}

/**
 * Initialize
 */
template <class TFixedImage, class TMovingImage>
void NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::Initialize(void) throw (ExceptionObject)
{
	this->Superclass::Initialize();

	m_NumberOfParameters = this->m_Transform->GetNumberOfParameters();
	m_NumberOfSpatialSamples = 
              this->GetFixedImageRegion().GetNumberOfPixels();

	/**
	 * Allocate memory for the fixed image sample container.
	 */
	m_FixedImageSamples.resize( m_NumberOfSpatialSamples );

	/** 
     * Take all the pixels within the fixed image region)
     * to create the sample points list.
     */
	this->SampleFullFixedImageDomain( m_FixedImageSamples );

	/**
	 * Check if the transform is of type BSplineDeformableTransform.
	 *
	 * If so, several speed up features are implemented.
	 * [1] Precomputing the results of bulk transform for each sample point.
	 * [2] Precomputing the BSpline weights for each sample point,
	 *     to be used later to directly compute the deformation vector
	 * [3] Precomputing the indices of the parameters within the
	 *     support region of each sample point.
	 */
	m_TransformIsBSpline = true;

	BSplineTransformType * testPtr = dynamic_cast<BSplineTransformType *>(
			this->m_Transform.GetPointer() );
	if( !testPtr )
	{
		m_TransformIsBSpline = false;
		m_BSplineTransform = NULL;
		itkDebugMacro( "Transform is not BSplineDeformable" );
	}
	else
	{
		m_BSplineTransform = testPtr;
		m_NumParametersPerDim =
			m_BSplineTransform->GetNumberOfParametersPerDimension();
		m_NumBSplineWeights = m_BSplineTransform->GetNumberOfWeights();
		itkDebugMacro( "Transform is BSplineDeformable" );
	}

	if( this->m_TransformIsBSpline )
	{
		// First, deallocate memory that may have been used from previous run of the Metric
		this->m_BSplineTransformWeightsArray.SetSize( 1, 1 );
		this->m_BSplineTransformIndicesArray.SetSize( 1, 1 );
		this->m_PreTransformPointsArray.resize( 1 );
		this->m_WithinSupportRegionArray.resize( 1 );
		this->m_Weights.SetSize( 1 );
		this->m_Indices.SetSize( 1 );

		if( this->m_UseCachingOfBSplineWeights )
		{
			m_BSplineTransformWeightsArray.SetSize( 
					m_NumberOfSpatialSamples, m_NumBSplineWeights );
			m_BSplineTransformIndicesArray.SetSize(
					m_NumberOfSpatialSamples, m_NumBSplineWeights );
			m_PreTransformPointsArray.resize( m_NumberOfSpatialSamples );
			m_WithinSupportRegionArray.resize( m_NumberOfSpatialSamples );

			this->PreComputeTransformValues();
		}
		else
		{
			this->m_Weights.SetSize( this->m_NumBSplineWeights );
			this->m_Indices.SetSize( this->m_NumBSplineWeights );
		}
		
		for ( unsigned int j = 0; j < FixedImageDimension; j++ )
		{
			m_ParametersOffset[j] = j *
				m_BSplineTransform->GetNumberOfParametersPerDimension();
		}
	}
}

/**
 * Sample the fixed image domain using all pixels in the Fixed image region
 */
template < class TFixedImage, class TMovingImage >
void
NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::SampleFullFixedImageDomain( FixedImageSpatialSampleContainer& samples )
{
  // Set up a region interator within the user specified fixed image region.
  typedef ImageRegionConstIteratorWithIndex<FixedImageType> RegionIterator;
  RegionIterator regionIter( this->m_FixedImage, this->GetFixedImageRegion() );

  regionIter.GoToBegin();

  typename FixedImageSpatialSampleContainer::iterator iter;
  typename FixedImageSpatialSampleContainer::const_iterator end=samples.end();

  if( this->m_FixedImageMask )
    {
    InputPointType inputPoint;

    iter=samples.begin();
    unsigned long nSamplesPicked = 0;

    while( iter != end && !regionIter.IsAtEnd() )
      {
      // Get sampled index
      FixedImageIndexType index = regionIter.GetIndex();
      // Check if the Index is inside the mask, translate index to point
      this->m_FixedImage->TransformIndexToPhysicalPoint( index, inputPoint );

      // If not inside the mask, ignore the point
      if( !this->m_FixedImageMask->IsInside( inputPoint ) )
        {
        ++regionIter; // jump to next pixel
        continue;
        }

      // Get sampled fixed image value
      (*iter).FixedImageValue = regionIter.Get();
      // Translate index to point
      (*iter).FixedImagePointValue = inputPoint;

      ++regionIter;
      ++iter;
      ++nSamplesPicked;
      }

    // If we picked fewer samples than the desired number, 
    // resize the container
    if (nSamplesPicked != this->m_NumberOfSpatialSamples)
      {
      this->m_NumberOfSpatialSamples = nSamplesPicked;
      samples.resize(this->m_NumberOfSpatialSamples);
      }
    }
  else // not restricting sample throwing to a mask
    {

    for( iter=samples.begin(); iter != end; ++iter )
      {
      // Get sampled index
      FixedImageIndexType index = regionIter.GetIndex();
      // Get sampled fixed image value
      (*iter).FixedImageValue = regionIter.Get();
      // Translate index to point
      this->m_FixedImage->TransformIndexToPhysicalPoint( index,
                                               (*iter).FixedImagePointValue );
      ++regionIter;
      }
    }
}

/**
 * Get the match Measure
 */
template <class TFixedImage, class TMovingImage> 
typename NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>::MeasureType
NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::GetValue( const TransformParametersType & parameters ) const
{

  FixedImageConstPointer fixedImage = this->m_FixedImage;

  if( !fixedImage ) 
    {
    itkExceptionMacro( << "Fixed image has not been assigned" );
    }

  MeasureType measure;

  this->m_NumberOfPixelsCounted = 0;

  this->SetTransformParameters( parameters );

  typedef  typename NumericTraits< MeasureType >::AccumulateType AccumulateType;

  AccumulateType sff = NumericTraits< AccumulateType >::Zero;
  AccumulateType smm = NumericTraits< AccumulateType >::Zero;
  AccumulateType sfm = NumericTraits< AccumulateType >::Zero;
  AccumulateType sf  = NumericTraits< AccumulateType >::Zero;
  AccumulateType sm  = NumericTraits< AccumulateType >::Zero;

  // Declare iterators for iteration over the sample container
  typename FixedImageSpatialSampleContainer::const_iterator fiter;
  typename FixedImageSpatialSampleContainer::const_iterator fend = 
    m_FixedImageSamples.end();

  unsigned long nFixedImageSamples=0;

  for ( fiter = m_FixedImageSamples.begin(); fiter != fend; ++fiter )
    {

	// Get moving image value
    MovingImagePointType mappedPoint;
    bool sampleOk;
    RealType movingValue;

    this->TransformPoint( nFixedImageSamples, parameters, mappedPoint, 
                          sampleOk, movingValue );

	if( sampleOk )
      {
	
	  const RealType fixedValue = m_FixedImageSamples[nFixedImageSamples].FixedImageValue;
	  sff += fixedValue  * fixedValue;
      smm += movingValue * movingValue;
	  sfm += fixedValue  * movingValue;
	  if ( this->m_SubtractMean )
        {
        sf += fixedValue;
        sm += movingValue;
        }
      this->m_NumberOfPixelsCounted++;
      }

	++nFixedImageSamples;
	}

  if ( this->m_SubtractMean && this->m_NumberOfPixelsCounted > 0 )
    {
    sff -= ( sf * sf / this->m_NumberOfPixelsCounted );
    smm -= ( sm * sm / this->m_NumberOfPixelsCounted );
    sfm -= ( sf * sm / this->m_NumberOfPixelsCounted );
    }

  const RealType denom = -1.0 * vcl_sqrt(sff * smm );

  if( this->m_NumberOfPixelsCounted > 0 && denom != 0.0)
    {
    measure = sfm / denom;
    }
  else
    {
    measure = NumericTraits< MeasureType >::Zero;
    }

  return measure;

}

/**
 * Get the Derivative Measure
 */
template < class TFixedImage, class TMovingImage> 
void
NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::GetDerivative( const TransformParametersType & parameters,
                 DerivativeType & derivative ) const
{
  MeasureType value;
  // call the combined version
  this->GetValueAndDerivative( parameters, value, derivative );
}


/*
 * Get both the match Measure and theDerivative Measure 
 */
template <class TFixedImage, class TMovingImage> 
void
NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::GetValueAndDerivative(const TransformParametersType & parameters, 
                        MeasureType & value, DerivativeType  & derivative) const
{


  if( !this->GetGradientImage() )
    {
    itkExceptionMacro(<<"The gradient image is null, maybe you forgot to call Initialize()");
    }

  FixedImageConstPointer fixedImage = this->m_FixedImage;

  if( !fixedImage ) 
    {
    itkExceptionMacro( << "Fixed image has not been assigned" );
    }

  const unsigned int dimension = FixedImageType::ImageDimension;

  this->m_NumberOfPixelsCounted = 0;

  this->SetTransformParameters( parameters );

  typedef  typename NumericTraits< MeasureType >::AccumulateType AccumulateType;

  AccumulateType sff  = NumericTraits< AccumulateType >::Zero;
  AccumulateType smm  = NumericTraits< AccumulateType >::Zero;
  AccumulateType sfm = NumericTraits< AccumulateType >::Zero;
  AccumulateType sf  = NumericTraits< AccumulateType >::Zero;
  AccumulateType sm  = NumericTraits< AccumulateType >::Zero;

  const unsigned int ParametersDimension = this->GetNumberOfParameters();
  derivative = DerivativeType( ParametersDimension );
  derivative.Fill( NumericTraits<ITK_TYPENAME DerivativeType::ValueType>::Zero );

  DerivativeType derivativeF = DerivativeType( ParametersDimension );
  derivativeF.Fill( NumericTraits<ITK_TYPENAME DerivativeType::ValueType>::Zero );

  DerivativeType derivativeM = DerivativeType( ParametersDimension );
  derivativeM.Fill( NumericTraits<ITK_TYPENAME DerivativeType::ValueType>::Zero );

  DerivativeType derivativeM1 = DerivativeType( ParametersDimension );
  derivativeM1.Fill( NumericTraits<ITK_TYPENAME DerivativeType::ValueType>::Zero );

  // Declare iterators for iteration over the sample container
  typename FixedImageSpatialSampleContainer::const_iterator fiter;
  typename FixedImageSpatialSampleContainer::const_iterator fend = 
    m_FixedImageSamples.end();

  unsigned long nFixedImageSamples=0;

  for ( fiter = m_FixedImageSamples.begin(); fiter != fend; ++fiter )
    {

	// Get moving image value
    MovingImagePointType mappedPoint;
    bool sampleOk;
    double movingValue;

	this->TransformPoint( nFixedImageSamples, parameters, mappedPoint, 
                          sampleOk, movingValue );

	if( sampleOk )
      {

	  const RealType fixedValue = m_FixedImageSamples[nFixedImageSamples].FixedImageValue;
      sff += fixedValue  * fixedValue;
      smm += movingValue * movingValue;
      sfm += fixedValue  * movingValue;
      if ( this->m_SubtractMean )
        {
        sf += fixedValue;
        sm += movingValue;
        }
      this->m_NumberOfPixelsCounted++;
      }

	++nFixedImageSamples;
	}

  // Compute contributions to derivatives
  nFixedImageSamples = 0;

  for ( fiter = m_FixedImageSamples.begin(); fiter != fend; ++fiter )
    {

	// Get moving image value
    MovingImagePointType mappedPoint;
    bool sampleOk;
    double movingValue;

	this->TransformPoint( nFixedImageSamples, parameters, mappedPoint, 
                          sampleOk, movingValue );

	if( sampleOk )
      {

      const RealType fixedValue = m_FixedImageSamples[nFixedImageSamples].FixedImageValue;
      const TransformJacobianType & jacobian =
		  this->m_Transform->GetJacobian(m_FixedImageSamples[nFixedImageSamples].FixedImagePointValue );

	  // Get the gradient by NearestNeighboorInterpolation: 
      // which is equivalent to round up the point components.
      typedef typename OutputPointType::CoordRepType CoordRepType;
      typedef ContinuousIndex<CoordRepType,MovingImageType::ImageDimension>
        MovingImageContinuousIndexType;

	  MovingImageContinuousIndexType tempIndex;
      this->m_MovingImage->TransformPhysicalPointToContinuousIndex( mappedPoint, tempIndex );

	  typename MovingImageType::IndexType mappedIndex; 
      mappedIndex.CopyWithRound( tempIndex );
      
      const GradientPixelType gradient = 
        this->GetGradientImage()->GetPixel( mappedIndex );

      for(unsigned int par=0; par<ParametersDimension; par++)
        {
        RealType sumF = NumericTraits< RealType >::Zero;
        RealType sumM = NumericTraits< RealType >::Zero;
        for(unsigned int dim=0; dim<dimension; dim++)
          {
          const RealType differential = jacobian( dim, par ) * gradient[dim];
          sumF += fixedValue  * differential;
          sumM += movingValue * differential;
          if ( this->m_SubtractMean && this->m_NumberOfPixelsCounted > 0 )
            {
            sumF -= differential * sf / this->m_NumberOfPixelsCounted;
            sumM -= differential * sm / this->m_NumberOfPixelsCounted;
            }
          }
        derivativeF[par] += sumF;
        derivativeM[par] += sumM;
        }
      }

	++nFixedImageSamples;
	}
      
  if ( this->m_SubtractMean && this->m_NumberOfPixelsCounted > 0 )
    {
    sff -= ( sf * sf / this->m_NumberOfPixelsCounted );
    smm -= ( sm * sm / this->m_NumberOfPixelsCounted );
    sfm -= ( sf * sm / this->m_NumberOfPixelsCounted );
    }

  const RealType denom = -1.0 * vcl_sqrt(sff * smm );

  if( this->m_NumberOfPixelsCounted > 0 && denom != 0.0)
    {
    for(unsigned int i=0; i<ParametersDimension; i++)
      {
      derivative[i] = ( derivativeF[i] - (sfm/smm)* derivativeM[i] ) / denom;
      }
    value = sfm / denom;
    }
  else
    {
    for(unsigned int i=0; i<ParametersDimension; i++)
      {
      derivative[i] = NumericTraits< MeasureType >::Zero;
      }
    value = NumericTraits< MeasureType >::Zero;
    }
}

template < class TFixedImage, class TMovingImage >
void
NormalizedCorrelationImageToImageMetricFast<TFixedImage, TMovingImage>
::TransformPoint(
  unsigned int sampleNumber,
  const TransformParametersType& parameters,
  MovingImagePointType& mappedPoint,
  bool& sampleOk,
  double& movingImageValue ) const
{
  
  if ( !m_TransformIsBSpline )
    {
    // Use generic transform to compute mapped position
    mappedPoint = this->m_Transform->TransformPoint( 
      m_FixedImageSamples[sampleNumber].FixedImagePointValue );

    // Check if mapped point inside image buffer
    sampleOk = this->m_Interpolator->IsInsideBuffer( mappedPoint );
    }
  else
	{
	
	if( this->m_UseCachingOfBSplineWeights )
      {  
      // If the transform is BSplineDeformable, we can use the precomputed
      // weights and indices to obtained the mapped position
      //
      const WeightsValueType * weights = 
                                   m_BSplineTransformWeightsArray[sampleNumber];
      const IndexValueType   * indices = 
                                   m_BSplineTransformIndicesArray[sampleNumber];
      mappedPoint.Fill( 0.0 );

      if ( m_WithinSupportRegionArray[sampleNumber] )
        {
        for ( unsigned int k = 0; k < m_NumBSplineWeights; k++ )
          {
          for ( unsigned int j = 0; j < FixedImageDimension; j++ )
            {
            mappedPoint[j] += weights[k] * 
              parameters[ indices[k] + m_ParametersOffset[j] ];
            }
          }
        }

      for( unsigned int j = 0; j < FixedImageDimension; j++ )
        {
        mappedPoint[j] += m_PreTransformPointsArray[sampleNumber][j];
        }

      // Check if mapped point inside image buffer
      sampleOk = this->m_Interpolator->IsInsideBuffer( mappedPoint );

      // Check if mapped point is within the support region of a grid point.
      // This is neccessary for computing the metric gradient
      sampleOk = sampleOk && m_WithinSupportRegionArray[sampleNumber];
      }
    else
      {
      // If not caching values, we invoke the Transform to recompute the
      // mapping of the point.
      this->m_BSplineTransform->TransformPoint( 
        this->m_FixedImageSamples[sampleNumber].FixedImagePointValue,
        mappedPoint, this->m_Weights, this->m_Indices, sampleOk);

      // Check if mapped point inside image buffer
      sampleOk = sampleOk && this->m_Interpolator->IsInsideBuffer( mappedPoint );
      }

    }

  // If user provided a mask over the Moving image
  if ( this->m_MovingImageMask )
    {
    // Check if mapped point is within the support region of the moving image
    // mask
    sampleOk = sampleOk && this->m_MovingImageMask->IsInside( mappedPoint );
    }


  if ( sampleOk )
    {
    movingImageValue = this->m_Interpolator->Evaluate( mappedPoint );
    }
}

/**
 * Cache pre-transformed points, weights and indices. 
 * This method is only called if the flag UseCachingOfBSplineWeights is ON.
 */
template < class TFixedImage, class TMovingImage >
void
NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::PreComputeTransformValues()
{
  // Create all zero dummy transform parameters
  ParametersType dummyParameters( this->m_Transform->GetNumberOfParameters() );
  dummyParameters.Fill( 0.0 );
  this->m_Transform->SetParameters( dummyParameters );

  // Cycle through each sampled fixed image point
  BSplineTransformWeightsType weights( m_NumBSplineWeights );
  BSplineTransformIndexArrayType indices( m_NumBSplineWeights );
  bool valid;
  MovingImagePointType mappedPoint;

  // Declare iterators for iteration over the sample container
  typename FixedImageSpatialSampleContainer::const_iterator fiter;
  typename FixedImageSpatialSampleContainer::const_iterator fend = 
    m_FixedImageSamples.end();
  unsigned long counter = 0;

  for( fiter = m_FixedImageSamples.begin(); fiter != fend; ++fiter, counter++ )
    {
    m_BSplineTransform->TransformPoint( 
                            m_FixedImageSamples[counter].FixedImagePointValue,
                            mappedPoint, weights, indices, valid );

    for( unsigned long k = 0; k < m_NumBSplineWeights; k++ )
      {
      m_BSplineTransformWeightsArray[counter][k] = weights[k];
      m_BSplineTransformIndicesArray[counter][k] = indices[k];
      }

    m_PreTransformPointsArray[counter]      = mappedPoint;
    m_WithinSupportRegionArray[counter]     = valid;

    }

}

template < class TFixedImage, class TMovingImage> 
void
NormalizedCorrelationImageToImageMetricFast<TFixedImage,TMovingImage>
::PrintSelf(std::ostream& os, Indent indent) const
{
  Superclass::PrintSelf(os, indent);
  os << indent << "SubtractMean: " << m_SubtractMean << std::endl;
  os << indent << "TransformIsBSpline: ";
  os << m_TransformIsBSpline << std::endl;
  os << indent << "UseCachingOfBSplineWeights: ";
  os << m_UseCachingOfBSplineWeights << std::endl;
  os << indent << "NumberOfSpatialSamples: ";
  os << m_NumberOfSpatialSamples << std::endl;
}

} // end namespace itk


#endif
